/*
   skipfish - main entry point
   ---------------------------

   Author: Michal Zalewski <lcamtuf@google.com>

   Copyright 2009, 2010 by Google Inc. All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>
#include <termios.h>
#include <fcntl.h>

#include <popt.h>
#define	N_(Text) Text	/* XXX FIXME: no i18n yet */

#include "types.h"
#include "alloc-inl.h"
#include "string-inl.h"

#include "crawler.h"
#include "analysis.h"
#include "database.h"
#include "http_client.h"
#include "report.h"

#ifdef DEBUG_ALLOCATOR
struct __AD_trk_obj* __AD_trk[ALLOC_BUCKETS];
u32 __AD_trk_cnt[ALLOC_BUCKETS];
#endif /* DEBUG_ALLOCATOR */

/* *BSD where J or Z is set are incompatible with our allocator. */
const char* malloc_options  = "jz";
const char* _malloc_options = "jz";

/* Ctrl-C handler... */

static u8 stop_soon, clear_screen;

static void ctrlc_handler(int sig) {
  stop_soon = 1;
}

/* Screen resizing handler. */

static void resize_handler(int sig) {
  clear_screen = 1;
}

/*==============================================================*/

#ifdef	NOTYET		/* XXX TODO: automate --option KEY=VAL,KEY=VAL,... */
static u8 * user_pass;
static u8 ** host_ip;		/* XXX multiple */
static u8 ** cookie_nv;		/* XXX multiple */
static u8 ** header_nv;		/* XXX multiple */
static u8 ** formhint_nv;	/* XXX multiple */
#endif

static u8 * browser;

static u32 seed;
static u32 purge_age;
static int purge_age_in = -1;	/* XXX option val must be >= 3 */
static int dont_save_words;
static int be_quiet;
static u8 * wordlist = (u8*)DEF_WORDLIST;
static const char * output_dir;

static struct poptOption skipfishAccessOptionsTable[] = {

  { NULL, 'A', POPT_ARG_STRING,	NULL, 'A',
	N_("- use specified HTTP authentication credentials"), N_(" user:pass") },
  { NULL, 'F', POPT_ARG_STRING,	NULL, 'F',
	N_("- pretend that 'host' resolves to 'IP'"), N_(" host=IP") },
  { NULL, 'C', POPT_ARG_STRING,	NULL, 'C',
	N_("- append a custom cookie to all requests"), N_(" name=val") },
  { NULL, 'H', POPT_ARG_STRING,	NULL, 'H',
	N_("- append a custom HTTP header to all requests"), N_(" name=val") },
  { NULL, 'b', POPT_ARG_STRING,	&browser, 0,
	N_("- use headers consistent with MSIE / Firefox"), N_(" (i|f)") },
  { NULL, 'N', POPT_ARG_VAL,	&ignore_cookies, 1,
	N_("- do not accept any new cookies"), NULL },

  POPT_TABLEEND
};

static struct poptOption skipfishCrawlOptionsTable[] = {

  { NULL, 'd', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_depth, 0,
	N_("- maximum crawl tree depth"), N_(" max_depth") },
  { NULL, 'c', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_children, 0,
	N_("- max total number of requests to send"), N_(" max_child") },
  { NULL, 'r', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_requests, 0,
	N_("- max total number of requests to send"), N_(" r_limit") },
  { NULL, 'p', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&crawl_prob, 0,
	N_("- node and link crawl probability"), N_(" crawl%") },
  { NULL, 'q', POPT_ARG_INT,	&seed, 0,
	N_("- repeat probabilistic scan with given seed"), N_(" hex") },
  { NULL, 'I', POPT_ARG_ARGV,	&allow_urls, 0,
	N_("- only follow URLs matching 'string'"), N_(" string") },
  { NULL, 'X', POPT_ARG_ARGV,	&deny_urls, 0,
	N_("- exclude URLs matching 'string'"), N_(" string") },
  { NULL, 'S', POPT_ARG_ARGV,	&deny_strings, 0,
	N_("- exclude pages containing 'string'"), N_(" string") },
  { NULL, 'K', POPT_ARG_ARGV,	&skip_params, 0,
	N_("- do not fuzz parameters named 'string'"), N_(" string") },
  { NULL, 'D', POPT_ARG_ARGV,	&allow_domains, 0,
	N_("- crawl cross-site links to another domain"), N_(" domain") },
  { NULL, 'B', POPT_ARG_ARGV,	&trust_domains, 0,
	N_("- trust, but do not crawl, another domain"), N_(" domain") },
  { NULL, 'Z', POPT_ARG_VAL,	&no_500_dir, 1,
	N_("- do not descend into 5xx locations"), NULL },
  { NULL, 'O', POPT_ARG_VAL,	&no_forms, 1,
	N_("- do not submit any forms"), NULL },
  { NULL, 'P', POPT_ARG_VAL,	&no_parse, 1,
	N_("- do not parse HTML, etc, to find new links"), NULL },

  POPT_TABLEEND
};

static struct poptOption skipfishReportOptionsTable[] = {

  { NULL, 'o', POPT_ARG_STRING,	&output_dir, 0,
	N_("- write output to specified directory (required)"), N_(" dir") },
  { NULL, 'J', POPT_ARG_VAL,	&relaxed_mime, 1,
	N_("- be less picky about MIME / charset mismatches"), NULL },
  { NULL, 'M', POPT_ARG_VAL,	&warn_mixed, 1,
	N_("- log warnings about mixed content"), NULL },
  { NULL, 'E', POPT_ARG_VAL,	&pedantic_cache, 1,
	N_("- log all HTTP/1.0 / HTTP/1.1 caching intent mismatches"), NULL },
  { NULL, 'U', POPT_ARG_VAL,	&log_ext_urls, 1,
	N_("- log all external URLs and e-mails seen"), NULL },
  { NULL, 'Q', POPT_ARG_VAL,	&suppress_dupes, 1,
	N_("- completely suppress duplicate nodes in reports"), NULL },
  { NULL, 'u', POPT_ARG_VAL,	&be_quiet, 1,
	N_("- be quiet, disable realtime progress stats"), NULL },

  POPT_TABLEEND
};

static struct poptOption skipfishDictOptionsTable[] = {

  { NULL, 'W', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT,	&wordlist, 0,
	N_("- load an alternative wordlist"), N_(" wordlist") },
  { NULL, 'L', POPT_ARG_VAL,	&dont_add_words, 1,
	N_("- do not auto-learn new keywords for the site"), NULL },
  { NULL, 'V', POPT_ARG_VAL,	&dont_save_words, 1,
	N_("- do not update wordlist based on scan results"), NULL },
  { NULL, 'Y', POPT_ARG_VAL,	&no_fuzz_ext, 1,
	N_("- do not fuzz extensions in directory brute-force"), NULL },
  { NULL, 'R', POPT_ARG_INT,	&purge_age_in, 0,
	N_("- purge words hit more than 'age' scans ago"), N_(" age") },
  { NULL, 'T', POPT_ARG_STRING,	NULL, 'T',
	N_("- add new form auto-fill rule"), N_(" name=val") },
  { NULL, 'G', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_guesses, 0,
	N_("- maximum number of keyword guesses to keep"), N_(" max_guess") },

  POPT_TABLEEND
};

static struct poptOption skipfishPerfOptionsTable[] = {

  { NULL, 'g', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_connections, 0,
	N_("- max simultaneous TCP connections, global"), N_(" max_conn") },
  { NULL, 'm', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_conn_host, 0,
	N_("- max simultaneous connections, per target IP"), N_(" host_conn") },
  { NULL, 'f', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&max_fail, 0,
	N_("- response size limit"), N_(" max_fail") },
  { NULL, 't', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&resp_tmout, 0,
	N_("- max number of consecutive HTTP errors"), N_(" req_tmout") },
  { NULL, 'w', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&rw_tmout, 0,
	N_("- total request response timeout"), N_(" rw_tmout") },
  { NULL, 'i', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&idle_tmout, 0,
	N_("- individual network I/O timeout"), N_(" idle_tmout") },
  { NULL, 's', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT,	&size_limit, 0,
	N_("- timeout on idle HTTP connections"), N_(" s_limit") },

  POPT_TABLEEND
};

static struct poptOption skipfishOptionsTable[] = {

  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, skipfishAccessOptionsTable, 0,
	N_("Authentication and access options:\n"), NULL },

  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, skipfishCrawlOptionsTable, 0,
	N_("Crawl scope options:\n"), NULL },

  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, skipfishReportOptionsTable, 0,
	N_("Reporting options:\n"), NULL },

  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, skipfishDictOptionsTable, 0,
	N_("Dictionary management options:\n"), NULL },

  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, skipfishPerfOptionsTable, 0,
	N_("Performance settings:\n"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP

  { NULL, (char)-1, POPT_ARG_INCLUDE_TABLE, NULL, 0,
        N_("\
Usage: %s [ options ... ] -o output_dir start_url [ start_url2 ... ]\n\
\n\
Send comments and complaints to <lcamtuf@google.com>.\n\
"), NULL },

  POPT_TABLEEND
};

/* Main entry point */

int main(int argc, char** argv) {
  u32 loop_cnt = 0;
  u8 show_once = 0, display_mode = 0;

  struct termios term;
  struct timeval tv;
  u64 st_time, en_time;

  const char * argv0;
  poptContext con;
  u8 ** av;
  int ac;
  int ec = 0;

  unsetenv("MALLOC_CHECK_");

  signal(SIGINT, ctrlc_handler);
  signal(SIGWINCH, resize_handler);
  signal(SIGPIPE, SIG_IGN);
  SSL_library_init();

  /* Come up with a quasi-decent random seed. */

  if (!seed) {
    gettimeofday(&tv, NULL);
    seed = tv.tv_usec ^ (tv.tv_sec << 16) ^ getpid();
  }

  SAY("skipfish version " VERSION " by <lcamtuf@google.com>\n");

  if ((argv0 = strrchr(argv[0], '/')) == NULL)
    argv0 = argv[0];
  else
    argv0++;

  con = poptGetContext(argv0, argc, (const char **) argv,
		skipfishOptionsTable, POPT_CONTEXT_POSIXMEHARDER);
  /* XXX FIXME: poptSetOtherOptionHelp() for custom Usage: msg */
  (void) poptReadDefaultConfig(con, 1);

  while ((ec = poptGetNextOpt(con)) > 0) {
    char * arg = poptGetOptArg(con);

    switch (ec) {

      case 'A': {
          u8* x = (u8*)strchr(arg, ':');
          if (!x) FATAL("Credentials must be in 'user:pass' form.");
          *(x++) = 0;
          auth_user = (u8*)arg;
          auth_pass = x;
          auth_type = AUTH_BASIC;
          break;
        }

      case 'F': {
          u8* x = (u8*)strchr(arg, '=');
          u32 fake_addr;
          if (!x) FATAL("Fake mappings must be in 'host=IP' form.");
          *x = 0;
          fake_addr = inet_addr((char*)x + 1);
          if (fake_addr == (u32)-1)
            FATAL("Could not parse IP address '%s'.", x + 1);
          fake_host((u8*)arg, fake_addr);
          break;
        }

      case 'H': {
          u8* x = (u8*)strchr(arg, '=');
          if (!x) FATAL("Extra headers must be in 'name=value' form.");
          *x = 0;
          if (!strcasecmp(arg, "Cookie"))
            FATAL("Do not use -H to set cookies (try -C instead).");
          SET_HDR((u8*)arg, x + 1, &global_http_par);
          break;
        }

      case 'C': {
          u8* x = (u8*)strchr(arg, '=');
          if (!x) FATAL("Cookies must be in 'name=value' form.");
          if (strchr(arg, ';'))
            FATAL("Split multiple cookies into separate -C options.");
          *x = 0;
          SET_CK((u8*)arg, x + 1, &global_http_par);
          break;
        }

      case 'T': {
          u8* x = (u8*)strchr(arg, '=');
          if (!x) FATAL("Rules must be in 'name=value' form.");
          *x = 0;
          add_form_hint((u8*)arg, x + 1);
          break;
        }
      default:
          FATAL("Option table is misconfigured.");
          ec = 1;
          exit(ec);
          break;
    }
    if (arg) free(arg);

  }

  if (ec < -1)
    FATAL("%s: %s: %s", argv0, poptBadOption(con, POPT_BADOPTION_NOALIAS),
      poptStrerror(ec));
  ec = 0;

  av = (u8 **) poptGetArgs(con);
  if (!(av && av[0])) {
    poptPrintUsage(con, stderr, 0);
    ec = 1;
    exit(ec);
  }

  if (allow_domains)	/* XXX FIXME: skip '*' as 1st char */
  while (allow_domains[num_allow_domains])
    num_allow_domains++;

  if (skip_params)	/* XXX FIXME: skip '*' as 1st char */
  while (skip_params[num_skip_params])
    num_skip_params++;

  if (trust_domains)	/* XXX FIXME: skip '*' as 1st char */
  while (trust_domains[num_trust_domains])
    num_trust_domains++;

  if (allow_urls)	/* XXX FIXME: skip '*' as 1st char */
  while (allow_urls[num_allow_urls])
    num_allow_urls++;

  if (deny_strings)	/* XXX FIXME: skip '*' as 1st char */
  while (deny_strings[num_deny_strings])
    num_deny_strings++;

  if (deny_urls)	/* XXX FIXME: skip '*' as 1st char */
  while (deny_urls[num_deny_urls])
    num_deny_urls++;

  if (purge_age_in >= 0) {
    purge_age = purge_age;
    if (purge_age < 3) FATAL("Purge age invalid or too low (min 3).");
  }
  if (max_depth < 2) FATAL("Invalid max_depth.");
  if (!max_children) FATAL("Invalid max_children.");
  if (!crawl_prob) FATAL("Invalid crawl_prob.");

  if (browser) {
    if (browser[0] == 'i') browser_type = BROWSER_MSIE; else
    if (browser[0] == 'f') browser_type = BROWSER_FFOX; else
      poptPrintHelp(con, stderr, 0);
      ec = 1;
      exit(ec);
  }

  if (!max_connections) FATAL("Invalid max_connections.");
  if (!max_conn_host) FATAL("Invalid max_conn_host.");
  if (!max_guesses) FATAL("Invalid max_guesses.");
  if (!max_requests) FATAL("Invalid max_requests.");
  if (!max_fail) FATAL("Invalid max_fail.");
  if (!resp_tmout) FATAL("Invalid resp_tmout.");
  if (!rw_tmout) FATAL("Invalid rw_tmout.");
  if (!idle_tmout) FATAL("Invalid idle_tmout.");
  if (!size_limit) FATAL("Invalid size_limit.");

  if (access(ASSETS_DIR "/index.html", R_OK))
    PFATAL("Unable to access '%s/index.html' - wrong directory?", ASSETS_DIR);

  srandom(seed);

  if (!output_dir)
    FATAL("Output directory not specified (try --help).");
  else {
    rmdir(output_dir);
    if (mkdir(output_dir, 0755))
      PFATAL("Unable to create '%s'.", output_dir);
  }

  if (resp_tmout < rw_tmout) 
    resp_tmout = rw_tmout;

  if (max_connections < max_conn_host)
    max_connections = max_conn_host;

  load_keywords((u8*)wordlist, purge_age);

  /* Schedule all URLs in the command line for scanning */

  ac = 0;
  while (av[ac]) {

    struct http_request *req = ck_alloc(sizeof(struct http_request));

    if (parse_url((u8*)av[ac], req, NULL))
      FATAL("One of specified scan targets is not a valid absolute URL.");

    if (!url_allowed_host(req))
      APPEND_FILTER(allow_domains, num_allow_domains,
                    __DFL_ck_strdup(req->host));

    if (!url_allowed(req))
      FATAL("URL '%s' explicitly excluded by -I / -X rules.", av[ac]);

    maybe_add_pivot(req, NULL, 2);
    destroy_request(req);

    ac++;
  }

  /* Char-by char stdin. */

  tcgetattr(0, &term);
  term.c_lflag &= ~ICANON;
  tcsetattr(0, TCSANOW, &term);
  fcntl(0, F_SETFL, O_NONBLOCK);

  gettimeofday(&tv, NULL);
  st_time = tv.tv_sec * 1000L + tv.tv_usec / 1000L;

  if (!be_quiet) SAY("\x1b[H\x1b[J");
    else SAY(cLGN "[*] " cBRI "Scan in progress, please stay tuned...\n");

  while ((next_from_queue() && !stop_soon) || (!show_once++)) {

    u8 keybuf[8];

    if (be_quiet || ((loop_cnt++ % 100) && !show_once)) continue;

    if (clear_screen) {
      SAY("\x1b[H\x1b[2J");
      clear_screen = 0;
    }

    SAY(cYEL "\x1b[H"
           "skipfish version " VERSION " by <lcamtuf@google.com>\n\n"
           cBRI "  -" cPIN " %s " cBRI "-\n\n" cNOR, 
           allow_domains[0]);


    if (!display_mode) {
      http_stats(st_time);
      SAY("\n");
      database_stats();
    } else {
      http_req_list();
    }

    SAY("        \r");

    if (fread(keybuf, 1, sizeof(keybuf), stdin) > 0) {
      display_mode ^= 1;
      clear_screen = 1;
    }

  }

  gettimeofday(&tv, NULL);
  en_time = tv.tv_sec * 1000L + tv.tv_usec / 1000L;

  SAY("\n");

  if (stop_soon)
    SAY(cYEL "[!] " cBRI "Scan aborted by user, bailing out!" cNOR "\n");

  term.c_lflag |= ICANON;
  tcsetattr(0, TCSANOW, &term);
  fcntl(0, F_SETFL, O_SYNC);

  if (!dont_save_words) save_keywords((u8*)wordlist);

  write_report((u8 *)output_dir, en_time - st_time, seed);

#ifdef LOG_STDERR
  SAY("\n== PIVOT DEBUG ==\n");
  dump_pivots(0, 0);
  SAY("\n== END OF DUMP ==\n\n");
#endif /* LOG_STDERR */

  SAY(cLGN "[+] " cBRI "This was a great day for science!" cRST "\n\n");

  if (!stop_soon) {
    destroy_database();
    destroy_http();
    destroy_signatures();
#ifdef DEBUG_ALLOCATOR
    __AD_report();
#endif /* DEBUG_ALLOCATOR */
  }

  EVP_cleanup();
  CRYPTO_cleanup_all_ex_data();

  fflush(0);

  con = poptFreeContext(con);

  return ec;

}
